/*----------------------------------------------------------------------------/
/  20-pin SD Sound Generator R0.01a                                           /
/-----------------------------------------------------------------------------/
/ This project, program codes and circuit diagrams, is opened under license
/ policy of following trems.
/
/  Copyright (C) 2013, ChaN, all right reserved.
/
/ * This project is a free software and there is NO WARRANTY.
/ * No restriction on use. You can use, modify and redistribute it for
/   personal, non-profit or commercial products UNDER YOUR RESPONSIBILITY.
/ * Redistributions of source code must retain the above copyright notice.
/
/----------------------------------------------------------------------------*/

// DigiGong - digital door bell by doc-diy, 2015
//
// program with:
// avrdude -c avrispmkII -p t861 -P usb -U flash:w:sd20p_hr.hex -U  lfuse:w:0xC1:m  -U hfuse:w:0xDF:m



#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/sleep.h>
#include <avr/wdt.h>
#include "pff.h"

#ifndef MODE
#error Wrong make file.
#endif

/*         tiny861
  SW0  PB0 |1   20| PA0  SD_DAT0
  OUT  PB1 |2   19| PA1  SD_CMD
  SW1  PB2 |3   18| PA2  SD_CLK
  OUT  PB3 |4   17| PA3  SD_DAT3
       VCC |5   16| GND
       GND |6   15| VCC
       PB4 |7   14| PA4
       PB5 |8   13| PA5
       PB6 |9   12| PA6  SHUTDOWN SD+AMP
       PB7 |10  11| PA7  LED
            ------

*/
//FUSES = {0xC1, 0xDD, 0xFF};	
/* ATtiny861 fuse bytes: Low, High, Extended.
This is the fuse settings of this project. The fuse data will be included
in the output hex file with program code. However some old flash programmers
cannot load the fuse bits from hex file. If it is the case, remove this line
and use these values to program the fuse bits. */

#define FCC(c1,c2,c3,c4)	(((DWORD)c4<<24)+((DWORD)c3<<16)+((WORD)c2<<8)+(BYTE)c1)	/* FourCC */

#define LED_ON()	PORTA |= _BV(7)
#define LED_OFF()	PORTA &= ~_BV(7)


#define BUTTONTHRESHHOLD   90 // 
#define BUTTONREPEATTIME   16 // max time between button presses

#define MAXPRESSESREAD     10 // maximum number of presses the button reading routine returns
#define LONGPRESS          18 // thershold for long press in 1/10 s
#define MAXSOUNDSGONG       3 // number of sounds played by gong

void delay_ms (WORD);	/* Defined in asmfunc.S */
void delay_us (WORD);	/* Defined in asmfunc.S */

EMPTY_INTERRUPT(PCINT_vect);


/*---------------------------------------------------------*/
/* Work Area                                               */
/*---------------------------------------------------------*/

volatile BYTE FifoRi, FifoWi, FifoCt;	/* FIFO controls */
BYTE Buff[256];		/* Audio output FIFO */

//BYTE InMode, Cmd;	/* Input mode and received command value */
BYTE Cmd;	/* Input mode and received command value */

FATFS Fs;			/* File system object */
DIR Dir;			/* Directory object */
FILINFO Fno;		/* File information */

WORD rb;			/* Return value. Put this here to avoid avr-gcc's bug */



/*---------------------------------------------------------*/
/* Sub-routines                                            */
/*---------------------------------------------------------*/


static
void led_sign (
	BYTE ct		/* Number of flashes */
)
{
	do {
		delay_ms(200);
		LED_ON();
		delay_ms(100);
		LED_OFF();
	} while (--ct);
	delay_ms(1000);
}



static
void ramp (		/* Ramp-up/down audio output (anti-pop feature) */
	int dir		/* 0:Ramp-down, 1:Ramp-up */
)
{
#if MODE != 0	/* This function is enebled on non-OCL output cfg. */
	BYTE v, d, n;


	if (dir) {
		v = 0; d = 1;
	} else {
		v = 128; d = 0xFF;
	}

	n = 128;
	do {
		v += d;
		OCR1A = v; OCR1B = v;
		delay_us(100);
	} while (--n);
#endif
}



static
void audio_on (void)	/* Enable audio output functions */
{
	if (!TCCR0B) {
		FifoCt = 0; FifoRi = 0; FifoWi = 0;		/* Reset audio FIFO */
		PLLCSR = 0b00000110;	/* Select PLL clock for TC1.ck */
		TCCR1A = 0b10100011;	/* Start TC1 with OC1A/OC1B PWM enabled */
		TCCR1B = 0b00000001;
		ramp(1);				/* Ramp-up to center level */
		TCCR0A = 0b00000001;	/* Enable TC0.ck = 2MHz as interval timer */
		TCCR0B = 0b00000010;
		TIMSK = _BV(OCIE0A);
	}
}



static
void audio_off (void)	/* Disable audio output functions */
{
	if (TCCR0B) {
		TCCR0B = 0;				/* Stop audio timer */
		ramp(0);				/* Ramp-down to GND level */
		TCCR1A = 0;	TCCR1B = 0;	/* Stop PWM */
	}
}




static
DWORD load_header (void)	/* 2:I/O error, 4:Invalid file, >=1024:Ok(number of samples) */
{
	DWORD sz, f;
	BYTE b, al = 0;


	/* Check RIFF-WAVE file header */
	if (pf_read(Buff, 12, &rb)) return 2;
	if (rb != 12 || LD_DWORD(Buff+8) != FCC('W','A','V','E')) return 4;

	for (;;) {
		if (pf_read(Buff, 8, &rb)) return 2;		/* Get Chunk ID and size */
		if (rb != 8) return 4;
		sz = LD_DWORD(&Buff[4]);		/* Chunk size */

		switch (LD_DWORD(&Buff[0])) {	/* Switch by chunk type */
		case FCC('f','m','t',' ') :		/* 'fmt ' chunk */
			if (sz & 1) sz++;
			if (sz > 128 || sz < 16) return 4;		/* Check chunk size */
			if (pf_read(Buff, sz, &rb)) return 2;	/* Get the chunk content */
			if (rb != sz) return 4;
			if (Buff[0] != 1) return 4;				/* Check coding type (1: LPCM) */
			b = Buff[2];
			if (b < 1 && b > 2) return 4; 			/* Check channels (1/2: Mono/Stereo) */
			GPIOR0 = al = b;						/* Save channel flag */
			b = Buff[14];
			if (b != 8 && b != 16) return 4;		/* Check resolution (8/16 bit) */
			GPIOR0 |= b;							/* Save resolution flag */
			if (b & 16) al <<= 1;
			f = LD_DWORD(&Buff[4]);					/* Check sampling freqency (8k-48k) */
			if (f < 8000 || f > 48000) return 4;
			OCR0A = (BYTE)(16000000UL/8/f) - 1;		/* Set interval timer (sampling period) */
			break;

		case FCC('d','a','t','a') :		/* 'data' chunk (start to play) */
			if (!al) return 4;							/* Check if format valid */
			if (sz < 1024 || (sz & (al - 1))) return 4;	/* Check size */
			if (Fs.fptr & (al - 1)) return 4;			/* Check offset */
			return sz;

		case FCC('D','I','S','P') :		/* 'DISP' chunk (skip) */
		case FCC('f','a','c','t') :		/* 'fact' chunk (skip) */
		case FCC('L','I','S','T') :		/* 'LIST' chunk (skip) */
			if (sz & 1) sz++;
			if (pf_lseek(Fs.fptr + sz)) return 2;
			break;

		default :						/* Unknown chunk */
			return 4;
		}
	}
}



static
BYTE play (		/* 0:Normal end, 1:Continue to play, 2:Disk error, 3:No file, 4:Invalid file */
	BYTE fn		/* File number (1..255) */
)
{
	DWORD sz, spa, sza;
	FRESULT res;
	WORD btr;
	BYTE n, i, rc;


//	if (InMode >= 2) Cmd = 0;	/* Clear command code (Edge triggered) */
Cmd = 0;
	/* Open an audio file "nnn.WAV" (nnn=001..255) */
	i = 2; n = fn;
	do {
		Buff[i] = (BYTE)(n % 10) + '0'; n /= 10;
	} while (i--);
	strcpy_P((char*)&Buff[3], PSTR(".WAV"));
	res = pf_open((char*)Buff);
	if (res == FR_NO_FILE) return 3;
	if (res != FR_OK) return 2;

	/* Get file parameters */
	sz = load_header();
	if (sz <= 4) return (BYTE)sz;	/* Invalid format */
	spa = Fs.fptr; sza = sz;		/* Save offset and size of audio data */

	LED_ON();
	audio_on();		/* Enable audio output */

	for (;;) {
		if (pf_read(0, 512 - (Fs.fptr % 512), &rb) != FR_OK) {		/* Snip sector unaligned part */
			rc = 2; break;
		}
		sz -= rb;
		do {
			/* Forward a bunch of the audio data to the FIFO */
			btr = (sz > 1024) ? 1024 : (WORD)sz;
			pf_read(0, btr, &rb);
			if (btr != rb) {
				rc = 2; break;
			}
			sz -= rb;

			/* Check input code change */
			rc = 0;
			wdt_reset();
		} while (!rc && rb == 1024);	/* Repeat until all data read or code change */

		if (rc || !Cmd) break;
	}

	while (FifoCt) ;			/* Wait for audio FIFO empty */
	OCR1A = 0x80; OCR1B = 0x80;	/* Return DAC out to center */

	LED_OFF();

	return rc;
}


void reset(void)
{
	wdt_enable(WDTO_15MS); // set WDE to generate chip reset (WDIE is insignificant here)
	set_sleep_mode(SLEEP_MODE_PWR_DOWN);
	while(1);
	{
		sleep_mode();      // save energy while waiting for reset (15ms)
	}
}


// read button value
int button(int n)
{
    if(n == 0) // input A
    {
        if(~PINB & (1<<PB0))
        {
            return 1;
        }
        else
        {
            return 0;
        }	
    }
    else       // input B
    {
        if(~PINB & (1<<PB2))
        {
            return 1;
        }
        else
        {
            return 0;
        }	
    }
}

// read buttons, detect multiple presses, robust to bounce and periodic gaps
// from an ac opto-coupler.
//
// general function: poll buttons 256 times and add 0 for depressed and 1 for 
// pressed to a variable. if the sum is higher than a threshold, the button is 
// detected as pressed.
void readbuttons(BYTE *pressesA, BYTE *pressesB, BYTE *pressTimeA, BYTE *pressTimeB)
{
    BYTE i;
    BYTE accuButtonA, accuButtonB; 
    BYTE pollPeriodsLeft;
    BYTE stateButtonAnew, stateButtonBnew;
    BYTE stateButtonAold = 0, stateButtonBold = 0;
    BYTE pressedPeriodsA = 0, pressedPeriodsB = 0;

    *pressesA = 0;
    *pressesB = 0;

    pollPeriodsLeft = 1;

    while(1) // loop over poll periods
    {
	accuButtonA = 0;
	accuButtonB = 0;

	for (i=0;i<255;i++) // 255 polls is one poll period
	{
		// read buttons and increase accumulator by 1 or 0
		accuButtonA = accuButtonA + button(0);
		accuButtonB = accuButtonB + button(1);
		// wait
		delay_us(392); // 2.55 kHz polling for 0.1 s poll periods, 25 polls per 1 rectified 50 Hz period
	} 

 	stateButtonAnew = (accuButtonA > BUTTONTHRESHHOLD); // check threshold
 	stateButtonBnew = (accuButtonB > BUTTONTHRESHHOLD);

        if(stateButtonAnew | stateButtonBnew) // if any button pressed
        {
            pollPeriodsLeft = BUTTONREPEATTIME; // go on with polling for next N periods
        }
	else // no button pressed
	{
	    pollPeriodsLeft--;
	}

        // capture button press duration
        if(stateButtonAnew)
	{
              if(pressedPeriodsA<255) pressedPeriodsA++;
        }
        if(stateButtonBnew) 
        {
              if(pressedPeriodsB<255) pressedPeriodsB++;
        }


        // capture button release, look for transition pressed->released
        if (stateButtonAold > stateButtonAnew)
        {
            if (*pressesA<MAXPRESSESREAD) (*pressesA)++;
            // record press time for button A
            *pressTimeA = pressedPeriodsA;
            pressedPeriodsA = 0;
        }
        if (stateButtonBold > stateButtonBnew)
        {
            if (*pressesB<MAXPRESSESREAD) (*pressesB)++;
            // record press time for button B
            *pressTimeB = pressedPeriodsB;
            pressedPeriodsB = 0;
        }


        // exit loop if no button action for last N cycles (timeout)
        if ( pollPeriodsLeft == 0)
        {
            break; // exit
        }

        stateButtonAold = stateButtonAnew;
        stateButtonBold = stateButtonBnew;
    }
}




/*-----------------------------------------------------------------------*/
/* Main                                                                  */

int main (void)
{
	BYTE rc, i;
        BYTE presses;
        BYTE pressesA, pressesB;
        BYTE pressTimeA, pressTimeB;

        MCUSR = 0;  /* Clear reset status */

	WDTCR = _BV(WDE) | _BV(WDCE);	/* Disable WDT */
	WDTCR = 0;

	/* Select pin change interrupt pins, 1: on, 0: off */
	PCMSK0 = 0;
 	PCMSK1 = (1<<PCINT10)|(1<<PCINT8);		/* PB0, PB2 */

	// set button pullups
	PORTA = 0b01000000;		/* LED and SHUTDOWN are outs, SD card lines "float"] */
	DDRA  = 0b11000000;             /* SHUTDOWN high disables sd card and amp */

	PORTB = 0b00000101;		/* pullups for switches, analog outs "float" */
	DDRB  = 0b00000000;

	// prepare sleep mode
	set_sleep_mode(SLEEP_MODE_PWR_DOWN);	/* Select power down mode for sleep */

	// wait in deep sleep for pin change, no watchdog running
	GIMSK = (1<<PCIE1)|(1<<PCIE0);   /* Enable pin change interrupt */
	sei();
	sleep_mode(); 
	GIMSK = 0;						/* Disable pin change interrupt */

	// enable sd card power, set command (number of wav file) needed for playback
	delay_ms(20);

	// read buttons
   	readbuttons(&pressesA, &pressesB, &pressTimeA, &pressTimeB);

	Cmd = 0;
	// select between door A and B based on number of presses
	if(pressesB>pressesA)
	{
		presses = pressesB;
		Cmd = Cmd + 100;                          // door B
        	if (pressTimeB>LONGPRESS) Cmd = Cmd + 10; // long press makes x0x.wav -> x1x.wav
	}
	else
	{
		presses = pressesA;
		Cmd = Cmd + 0;                            // door A
		if (pressTimeA>LONGPRESS) Cmd = Cmd + 10; // long press makes x0x.wav -> x1x.wav
	}

	if (presses>MAXSOUNDSGONG) presses = MAXSOUNDSGONG;         // limit presses to N (xxN.wav)

	Cmd = Cmd + presses;                        // number of presses sets least significant digit (starting with 0)


	// play sound

	/* Initialize ports, p: pull-up, L: low-out, H: high-out */
	PORTA = 0b00001011;	/* SHUTDOWN goes low (active), SD_CLK out, SD_CMD out, ... */
	DDRA  = 0b11001110;

	PORTB = 0b00000101;	/* analog outs become out */
	DDRB  = 0b00001010;


	//looks like wdt is needed for sleep with reduced current (about 7 uA)
	WDTCR = _BV(WDE) | 0b110;			/* Enable WDT (1s), action on time-out: reset */
	set_sleep_mode(SLEEP_MODE_PWR_DOWN);		/* Select power down mode for sleep */


	// make 5 attempts to read sd card
	for (i=0;i<5;i++) {
		if (pf_mount(&Fs) == FR_OK) {		/* Initialize FS */

			rc = play(Cmd);			/* Play corresponding audio file */
			if (rc >= 2) led_sign(rc);	/* Display if any error occured */

			audio_off();			/* Disable audio output */

                        reset();  // reset system and go to sleep
		}

		led_sign(2);	/* Disk error or Media mount failed */

	}
        //return 0;
}

